import csv, codecs

"""
Adapted from the csv examples:

http://docs.python.org/2/library/csv.html#csv-examples
"""


class UnicodeDictReader:
    """
    A UnicodeDictReader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding, decoding each field entry to
    unicode.
    """
    def __init__(self, f, dialect=csv.excel, fieldnames=None, restkey=None, restval=None, encoding="utf-8", *args, **kwds):
        self._encoding = encoding
        self._reader = csv.DictReader(f, dialect=dialect, fieldnames=fieldnames, restkey=restkey, restval=restval, *args, **kwds)

    def next(self):
        row = self._reader.next()
        return dict(zip(row.keys(), [codecs.decode(s, self._encoding) for s in row.values()]))

    def __iter__(self):
        return self


from django.conf.urls import patterns, url

from .views import FoodDetail, FoodList

urlpatterns = patterns('usda.views',
    url(r'^$', FoodList.as_view(), name='usda-food_list'),
    url(r'^(?P<ndb_number>\d+)/$', FoodDetail.as_view(), name='usda-food_detail'),
)



__author__ = 'jpalanca'

# Import the MongoAdmin base class
from mongonaut.sites import MongoAdmin

# Import your custom models
from .models import Food, FoodGroup, Weight, Nutrient, Footnote, DataDerivation, DataSource, Source, NutrientData

# Instantiate the MongoAdmin class
# Then attach the mongoadmin to your model
Food.mongoadmin = MongoAdmin()
FoodGroup.mongoadmin = MongoAdmin()
Weight.mongoadmin = MongoAdmin()
Nutrient.mongoadmin = MongoAdmin()
Footnote.mongoadmin = MongoAdmin()
DataSource.mongoadmin = MongoAdmin()
DataDerivation.mongoadmin = MongoAdmin()
Source.mongoadmin = MongoAdmin()
NutrientData.mongoadmin = MongoAdmin()



from mongoengine import Document, IntField, StringField, BooleanField, FloatField, ReferenceField, ListField


FOOTNOTE_DESC = 'D'
FOOTNOTE_MEAS = 'M'
FOOTNOTE_NUTR = 'N'

FOOTNOTE_CHOICES = (
    (FOOTNOTE_DESC, 'Footnote adding information to the food  description'),
    (FOOTNOTE_MEAS, 'Footnote adding information to measure description'),
    (FOOTNOTE_NUTR, 'Footnote providing additional information on a nutrient value'),
)


class Food(Document):
    ndb_number = IntField(verbose_name='Nutrient Databank Number', primary_key=True, help_text='Nutrient Databank number that uniquely identifies a food item.')
    food_group = ReferenceField('FoodGroup', verbose_name='Food Group', help_text='Food group to which a food item belongs.')
    long_description = StringField(verbose_name='Long Description', max_length=200, help_text='Description of food item')
    short_description = StringField(verbose_name='Short Description', max_length=60, help_text='Description of food item')
    common_name = StringField(verbose_name='Common Name', max_length=100, help_text='Other names commonly used to describe a food, including local or regional names for various foods, for example, "soda" or "pop" for "carbonated beverages."')
    manufacturer_name = StringField(verbose_name='Manufacturer Name', max_length=65, help_text='Indicates the company that manufactured the product, when appropriate.')
    survey = BooleanField(verbose_name='Survey', default=False, help_text='Indicates if the food item is used in the USDA Food and Nutrient Database for Dietary Studies (FNDDS) and thus has a complete nutrient profile for the 65 FNDDS nutrients.')
    refuse_description = StringField(verbose_name='Refuse Description', max_length=135, help_text='Description of inedible parts of a food item (refuse), such as seeds or bone.')
    refuse_percentage = IntField(verbose_name='Refuse Percentage', null=True, help_text='Percentage of refuse.')
    scientific_name = StringField(verbose_name='Scientific Name', max_length=65, help_text='Scientific name of the food item. Given for the least processed form of the food (usually raw), if applicable.')
    nitrogen_factor = FloatField(verbose_name='Nitrogen Factor', null=True, help_text='Factor for converting nitrogen to protein.')
    protein_factor = FloatField(verbose_name='Protein Factor', null=True, help_text='Factor for calculating calories from protein.')
    fat_factor = FloatField(verbose_name='Fat Factor', null=True, help_text='Factor for calculating calories from fat.')
    cho_factor = FloatField(verbose_name='CHO Factor', null=True, help_text='Factor for calculating calories from carbohydrate.')

    class Meta:
        verbose_name = 'Food'
        verbose_name_plural = 'Foods'
        ordering = ['ndb_number']
        db_alias = 'usda'

    def __unicode__(self):
        return self.long_description

    def get_absolute_url(self):
        return 'usda-food_detail', (), {'ndb_number': self.ndb_number}


class FoodGroup(Document):
    code = IntField(verbose_name='Food Group Code', primary_key=True, help_text='Code identifying a food group. Codes may not be consecutive.')
    description = StringField(verbose_name='Description', max_length=60, help_text='Name of food group.')
    
    class Meta:
        verbose_name = 'Food Group'
        verbose_name_plural = 'Food Groups'
        ordering = ['code']
        db_alias = 'usda'

    def __unicode__(self):
        return self.description


class Nutrient(Document):
    number = IntField(verbose_name='Number', primary_key=True, help_text='Unique identifier code for a nutrient.')
    units = StringField(verbose_name='Units', max_length=7, help_text='Units of measure (mg, g, and so on).')
    tagname = StringField(verbose_name='Tagname', max_length=20, help_text='International Network of Food Data Systems (INFOODS) Tagnames. A unique abbreviation for a nutrient/food component developed by INFOODS to aid in the interchange of data.')
    description = StringField(verbose_name='Description', max_length=60, help_text='Name of nutrient/food component.')
    decimals = IntField(verbose_name='Decimals', help_text='Number of decimal places to which a nutrient value is rounded.')
    order = IntField(verbose_name='Order', help_text='Used to sort nutrient records in the same order as various reports produced from SR.')

    class Meta:
        verbose_name = 'Nutrient'
        verbose_name_plural = 'Nutrients'
        ordering = ['number']
        db_alias = 'usda'
    
    def __unicode__(self):
        return self.description


class NutrientData(Document):
    food = ReferenceField('Food', verbose_name='Food')
    nutrient = ReferenceField('Nutrient', verbose_name='Nutrient')
    nutrient_value = FloatField(verbose_name='Nutrient Value', help_text='Amount in 100 grams, edible portion.')
    data_points = IntField(verbose_name='Data Points', help_text='Number of data points (previously called Sample_Ct) is the number of analyses used to calculate the nutrient value. If the number of data points is 0, the value was calculated or imputed.')
    standard_error = FloatField(verbose_name='Standard Error', null=True, help_text='Standard error of the mean. Null if can not be calculated.')
    source = ListField(ReferenceField('Source', verbose_name='Source', help_text='Type of data'))
    data_derivation = ReferenceField('DataDerivation', verbose_name='Data Derivation', null=True, help_text='Data Derivation giving specific information on how the value is determined.')
    reference_nbd_number = IntField(verbose_name='Reference NBD Number', null=True, help_text='NDB number of the item used to impute a missing value. Populated only for items added or updated starting with SR14.')
    added_nutrient = BooleanField(verbose_name='Added Nutritient', default=False, help_text='Indicates a vitamin or mineral added for fortification or enrichment. This field is populated for ready-to-eat breakfast cereals and many brand-name hot cereals in food group 8.')
    number_of_studies = IntField(verbose_name='Number of Studies', null=True, help_text='Number of studies.')
    minimum = FloatField(verbose_name='Minimum', null=True, help_text='Minimum value.')
    maximum = FloatField(verbose_name='Maximum', null=True, help_text='Maximum value.')
    degrees_of_freedom = IntField(verbose_name='Degrees of Freedom', null=True, help_text='Degrees of freedom.')
    lower_error_bound = FloatField(verbose_name='Lower Error Bound.', null=True, help_text='Lower 95% error bound.')
    upper_error_bound = FloatField(verbose_name='Upper Error Bound.', null=True, help_text='Upper 95% error bound.')
    statistical_comments = StringField(verbose_name='Statistical Comments', max_length=10, help_text='Statistical comments.')
    confidence_code = StringField(verbose_name='Confidence Code', max_length=1, help_text='Confidence Code indicating data quality, based on evaluation of sample plan, sample handling, analytical method, analytical quality control, and number of samples analyzed. Not included in this release, but is planned for future releases.')

    class Meta:
        verbose_name = 'Nutrient Data'
        verbose_name_plural = 'Nutrient Data'
        ordering = ['food', 'nutrient']
        unique_together = ['food', 'nutrient']
        db_alias = 'usda'

    def __unicode__(self):
        return u'%s - %s' % (self.food, self.nutrient)


class Source(Document):
    code = IntField(verbose_name='Code', primary_key=True)
    description = StringField(verbose_name='Description', max_length=60, help_text='Description of source code that identifies the type of nutrient data.')

    class Meta:
        verbose_name = 'Source'
        verbose_name_plural = 'Sources'
        ordering = ['code']
        db_alias = 'usda'

    def __unicode__(self):
        return self.description


class DataDerivation(Document):
    code = StringField(verbose_name='Code', max_length=4, primary_key=True)
    description = StringField(verbose_name='Description', max_length=120, help_text='Description of derivation code giving specific information on how the value was determined.')

    class Meta:
        verbose_name = 'Data Derivation'
        verbose_name_plural = 'Data Derivations'
        ordering = ['code']
        db_alias = 'usda'

    def __unicode__(self):
        return self.description


class Weight(Document):
    food = ReferenceField('Food', verbose_name='Food')
    sequence = IntField(verbose_name='Sequence', help_text='Sequence number.')
    amount = FloatField(verbose_name='Amount', help_text='Unit modifier (for example, 1 in "1 cup").')
    description = StringField(verbose_name='Description', max_length=80, help_text='Description (for example, cup, diced, and 1-inch pieces).')
    gram_weight = FloatField(verbose_name='Gram Weight', help_text='Gram weight.')
    number_of_data_points = FloatField(verbose_name='Number of Data Points', null=True, help_text='Number of data points.')
    standard_deviation = FloatField(verbose_name='Standard Deviation', null=True, help_text='Standard Deviation')

    class Meta:
        verbose_name = 'Weight'
        verbose_name_plural = 'Weights'
        ordering = ['food', 'sequence']
        unique_together = ['food', 'sequence']
        db_alias = 'usda'

    def __unicode__(self):
        return u'%d %s %s %dg' % (self.amount, self.description, self.food, self.gram_weight)


class Footnote(Document):
    food = ReferenceField('Food', verbose_name='Food')
    number = IntField(verbose_name='Sequence', help_text='Sequence number. If a given footnote applies to more than one nutrient number, the same footnote number is used. As a result, this file cannot be indexed.')
    type = StringField(verbose_name='Type', max_length=1, choices=FOOTNOTE_CHOICES, help_text='Type of footnote.')
    nutrient = ReferenceField('Nutrient', verbose_name='nutrient', null=True)
    text = StringField(verbose_name='Text', max_length=200, help_text='Footnote text.')
    
    class Meta:
        verbose_name = 'Footnote'
        verbose_name_plural = 'Footnotes'
        ordering = ['food', 'number']
        unique_together = ['food', 'number', 'nutrient']
        db_alias = 'usda'
    
    def __unicode__(self):
        return self.text


class DataSource(Document):
    id = StringField(verbose_name='Id', max_length=6, primary_key=True, help_text='Unique number identifying the reference/source.')
    authors = StringField(verbose_name='Authors', max_length=255, help_text='List of authors for a journal article or name of sponsoring organization for other documents.')
    title = StringField(verbose_name='Title', max_length=255, help_text='Title of article or name of document, such as a report from a company or trade association.')
    year = IntField(verbose_name='Year', null=True, help_text='Year article or document was published.')
    journal = StringField(verbose_name='Journal', max_length=135, null=True, help_text='Name of the journal in which the article was published.')
    volume_or_city = StringField(verbose_name='Volume or City', max_length=16, help_text='Volume number for journal articles, books, or reports; city where sponsoring organization is located.')
    issue_or_state = StringField(verbose_name='Issue or State', max_length=5, help_text='Issue number for journal article; State where the sponsoring organization is located.')
    start_page = IntField(verbose_name='Start Page', null=True, help_text='Starting page number of article/document.')
    end_page = IntField(verbose_name='End Page', null=True, help_text='Ending page number of article/document.')

    class Meta:
        verbose_name = 'Data Source'
        verbose_name_plural = 'Data Sources'
        ordering = ['id']
        db_alias = 'usda'
    
    def __unicode__(self):
        return self.title


import csv
import optparse
import logging
import os
import zipfile

from django.core.management.base import BaseCommand, CommandError

from usda_mongo.models import FOOTNOTE_DESC, FOOTNOTE_MEAS, FOOTNOTE_NUTR
from mongoengine.context_managers import switch_db
from .unicode_dict_reader import UnicodeDictReader


# Number of nutrient data items to read at once. Setting this too high may cause
# a MemoryException when `Debug=True` as query debugging information remains in RAM
NUTRIENT_DATA_STEP = 1000


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        optparse.make_option('-f', '--filename', action='store', dest='filename', help='The compressed SR22 filename', default='sr22.zip'),
        optparse.make_option('--database', action='store', dest='database', help='Specify database to load data into. Defaults to the "default" database.', default="default"),
        optparse.make_option('--all', action='store_true', dest='all', help='Create/Update all data.'),
        optparse.make_option('--food', action='store_true', dest='food', help='Create/Update foods.'),
        optparse.make_option('--group', action='store_true', dest='group', help='Create/Update food groups.'),
        optparse.make_option('--nutrient', action='store_true', dest='nutrient', help='Create/Update nutrients.'),
        optparse.make_option('--weight', action='store_true', dest='weight', help='Create/Update weights.'),
        optparse.make_option('--footnote', action='store_true', dest='footnote', help='Create/Update footnotes.'),
        optparse.make_option('--datasource', action='store_true', dest='datasource', help='Create/Update data sources.'),
        optparse.make_option('--derivation', action='store_true', dest='derivation', help='Create/Update data derivations.'),
        optparse.make_option('--source', action='store_true', dest='source', help='Create/Update sources.'),
        optparse.make_option('--data', action='store_true', dest='data', help='Create/Update nutrient data.'),
        optparse.make_option('--encoding', action='store', dest='encoding', help='Specify the src file encoding. Defaults to cp1252.', default='cp1252')
    )
    help = 'Updates/Created all SR22 data.'

    def handle(self, **options):
        FOOD_DES = 'FOOD_DES.txt'
        FD_GROUP = 'FD_GROUP.txt'
        NUT_DATA = 'NUT_DATA.txt'
        NUTR_DEF = 'NUTR_DEF.txt'
        SRC_CD = 'SRC_CD.txt'
        DERIV_CD = 'DERIV_CD.txt'
        WEIGHT = 'WEIGHT.txt'
        FOOTNOTE = 'FOOTNOTE.txt'
        DATSRCLN = 'DATSRCLN.txt'
        DATA_SRC = 'DATA_SRC.txt'

        required_files = [
            FOOD_DES,
            FD_GROUP,
            NUT_DATA,
            NUTR_DEF,
            SRC_CD,
            DERIV_CD,
            WEIGHT,
            FOOTNOTE,
            DATSRCLN,
            DATA_SRC,
        ]

        verbosity = int(options.get('verbosity', 1))
        db_alias = options.get('database', "default")
        parse_all = options.get('all')
        parse_group = options.get('group')
        parse_food = options.get('food')
        parse_weight = options.get('weight')
        parse_nutrient = options.get('nutrient')
        parse_footnote = options.get('footnote')
        parse_datasource = options.get('datasource')
        parse_derivation = options.get('derivation')
        parse_source = options.get('source')
        parse_data = options.get('data')
        encoding = options.get('encoding', 'cp1252')

        if not os.path.exists(options['filename']):
            CommandError('%s does not exist' % options['filename'])

        if verbosity == 1:
            logging.basicConfig(level=logging.INFO, format='%(levelname)s - %(message)s')
        elif verbosity > 1:
            logging.basicConfig(level=logging.DEBUG, format='%(levelname)s - %(message)s')

        logging.info('Verifying %s...' % options['filename'])

        if not parse_all and not any([
            parse_group, parse_food, parse_weight, parse_nutrient, parse_footnote,
            parse_datasource, parse_derivation, parse_source, parse_data
        ]):
            logging.info('Parsing all available data from %s' % options['filename'])
            parse_all = True

        zip_file = zipfile.ZipFile(options['filename'], mode='r')

        # Verify integrity of zip file by checking that all required files are present
        missing_files = [required_file for required_file in required_files if required_file not in zip_file.namelist()]
        if missing_files:
            logging.error('%s does not appear to be a valid SR22 database.  Unable to extract %s' % (options['filename'], ', '.join(missing_files)))

        if parse_all or parse_group:
            logging.info('Reading %s...' % FD_GROUP)
            create_update_food_groups(''.join([byte for byte in zip_file.read(FD_GROUP)]).splitlines(), db_alias)
        if parse_all or parse_food:
            logging.info('Reading %s...' % FOOD_DES)
            create_update_foods(''.join([byte for byte in zip_file.read(FOOD_DES)]).splitlines(), encoding, db_alias)
        if parse_all or parse_weight:
            logging.info('Reading %s...' % WEIGHT)
            create_update_weights(''.join([byte for byte in zip_file.read(WEIGHT)]).splitlines(), db_alias)
        if parse_all or parse_nutrient:
            logging.info('Reading %s...' % NUTR_DEF)
            create_update_nutrients(''.join([byte for byte in zip_file.read(NUTR_DEF)]).splitlines(), encoding, db_alias)
        if parse_all or parse_footnote:
            logging.info('Reading %s...' % FOOTNOTE)
            create_update_footnotes(''.join([byte for byte in zip_file.read(FOOTNOTE)]).splitlines(), db_alias)
        if parse_all or parse_datasource:
            logging.info('Reading %s...' % DATA_SRC)
            create_update_data_sources(''.join([byte for byte in zip_file.read(DATA_SRC)]).splitlines(), db_alias)
        if parse_all or parse_derivation:
            logging.info('Reading %s...' % DERIV_CD)
            create_update_derivations(''.join([byte for byte in zip_file.read(DERIV_CD)]).splitlines(), db_alias)
        if parse_all or parse_source:
            logging.info('Reading %s...' % SRC_CD)
            create_update_sources(''.join([byte for byte in zip_file.read(SRC_CD)]).splitlines(), db_alias)
        if parse_all or parse_data:
            logging.info('Reading %s...' % NUT_DATA)
            create_update_nutrient_data(''.join([byte for byte in zip_file.read(NUT_DATA)]).splitlines(), db_alias)

        zip_file.close()


def create_update_food_groups(data, db_alias):
    total_created = 0
    total_updated = 0

    logging.info('Processing %d food groups' % len(data))

    from usda_mongo.models import FoodGroup

    with switch_db(FoodGroup, db_alias) as FoodGroup:

        for row in csv.DictReader(
            data, fieldnames=('fdgrp_cd', 'fdgrp_desc'),
            delimiter='^', quotechar='~'
        ):
            created = False

            try:
                food_group = FoodGroup.objects.get(code=int(row['fdgrp_cd']))
                total_updated += 1
            except FoodGroup.DoesNotExist:
                food_group = FoodGroup(code=int(row['fdgrp_cd']))
                total_created += 1
                created = True

            food_group.description = row['fdgrp_desc']
            food_group.save()

            if created:
                logging.debug('Created %s' % food_group)
            else:
                logging.debug('Updated %s' % food_group)

        logging.info('Created %d new food groups' % total_created)
        logging.info('Updated %d food groups' % total_updated)


def create_update_foods(data, encoding, db_alias):
    total_created = 0
    total_updated = 0

    logging.info('Processing %d foods' % len(data))

    from usda_mongo.models import Food, FoodGroup

    with switch_db(Food, db_alias) as Food:
        with switch_db(FoodGroup, db_alias) as FoodGroup:

            for row in UnicodeDictReader(
                data,
                fieldnames=(
                    'ndb_no', 'fdgrp_cd', 'long_desc', 'short_desc', 'com_name',
                    'manufac_name', 'survey', 'ref_desc', 'refuse', 'sci_name',
                    'n_factor', 'pro_factor', 'fat_factor', 'cho_factor'
                ),
                delimiter='^', quotechar='~',
                encoding=encoding
            ):
                created = False

                try:
                    food = Food.objects.get(ndb_number=int(row['ndb_no']))
                    total_updated += 1
                except Food.DoesNotExist:
                    food = Food(ndb_number=int(row['ndb_no']))
                    total_created += 1
                    created = True

                food.food_group = FoodGroup.objects.get(code=int(row['fdgrp_cd']))
                food.long_description = row.get('long_desc')
                food.short_description = row.get('short_desc')
                food.common_name = row.get('com_name')
                food.manufacturer_name = row.get('manufac_name')
                if row.get('survey'):
                    food.survey = row['survey'] == 'Y'
                food.refuse_description = row.get('ref_desc')
                if row.get('refuse'):
                    food.refuse_precentage = int(row['refuse'])
                if row.get('sci_name'):
                    food.scientific_name = row['sci_name']
                if row.get('n_factor'):
                    food.nitrogen_factor = float(row['n_factor'])
                if row.get('pro_factor'):
                    food.protein_factor = float(row['pro_factor'])
                if row.get('fat_factor'):
                    food.fat_factor = float(row['fat_factor'])
                if row.get('cho_factor'):
                    food.cho_factor = float(row['cho_factor'])

                food.save()

                if created:
                    logging.debug('Created %s' % food)
                else:
                    logging.debug('Updated %s' % food)

    logging.info('Created %d new foods' % total_created)
    logging.info('Updated %d foods' % total_updated)


def create_update_weights(data, db_alias):
    total_created = 0
    total_updated = 0

    logging.info('Processing %d weights' % len(data))

    from usda_mongo.models import Weight, Food

    with switch_db(Weight, db_alias) as Weight:
        with switch_db(Food, db_alias) as Food:

            for row in csv.DictReader(
                data, fieldnames=(
                    'ndb_no', 'seq', 'amount', 'msre_desc', 'gm_wgt', 'num_data_pts', 'std_dev'
                ),
                delimiter='^', quotechar='~'
            ):
                created = False

                try:
                    weight = Weight.objects.get(
                        food=Food.objects.get(ndb_number=int(row['ndb_no'])),
                        sequence=int(row['seq'])
                    )
                    total_updated += 1
                except Weight.DoesNotExist:
                    weight = Weight(
                        food=Food.objects.get(ndb_number=int(row['ndb_no'])),
                        sequence=int(row['seq'])
                    )
                    total_created += 1
                    created = True

                weight.amount = float(row.get('amount'))
                weight.description = row.get('msre_desc')
                weight.gram_weight = float(row.get('gm_wgt'))
                if row.get('num_data_pts'):
                    weight.number_of_data_points = float(row['num_data_pts'])
                if row.get('std_dev'):
                    weight.standard_deviation = float(row['std_dev'])
                weight.save()

                if created:
                    logging.debug('Created %s' % weight)
                else:
                    logging.debug('Updated %s' % weight)

    logging.info('Created %d new weights' % total_created)
    logging.info('Updated %d weights' % total_updated)


def create_update_nutrients(data, encoding, db_alias):
    total_created = 0
    total_updated = 0

    logging.info('Processing %d nutrients' % len(data))

    from usda_mongo.models import Nutrient

    with switch_db(Nutrient, db_alias) as Nutrient:

        for row in UnicodeDictReader(
            data, fieldnames=(
                'nutr_no', 'units', 'tagname', 'nutrdesc', 'num_dec', 'sr_order'
            ),
            delimiter='^', quotechar='~',
            encoding=encoding
        ):
            created = False

            try:
                nutrient = Nutrient.objects.get(number=int(row['nutr_no']))
                total_updated += 1
            except Nutrient.DoesNotExist:
                nutrient = Nutrient(number=int(row['nutr_no']))
                total_created += 1
                created = True

            nutrient.units = row['units']
            nutrient.tagname = row.get('tagname')
            nutrient.description = row['nutrdesc']
            nutrient.decimals = int(row['num_dec'])
            nutrient.order = int(row['sr_order'])
            nutrient.save()

            if created:
                logging.debug('Created %s' % nutrient)
            else:
                logging.debug('Updated %s' % nutrient)

    logging.info('Created %d new nutrients' % total_created)
    logging.info('Updated %d nutrients' % total_updated)


def create_update_footnotes(data, db_alias):
    total_created = 0
    total_updated = 0

    logging.info('Processing %d footnotes' % len(data))

    from usda_mongo.models import Footnote, Nutrient, Food

    with switch_db(Footnote, db_alias) as Footnote:
        with switch_db(Nutrient, db_alias) as Nutrient:
            with switch_db(Food, db_alias) as Food:

                for row in csv.DictReader(
                    data, fieldnames=(
                        'ndb_no', 'footnt_no', 'footnt_typ', 'nutr_no', 'footnt_txt'
                    ),
                    delimiter='^', quotechar='~'
                ):
                    created = False

                    # SR22 definition indicates that `footnt_no` and `footnt_typ` are required,
                    # but on occasion, either on is blank.  To compensate for this, we assume
                    # a blank `footnt_no` is '1' and a blank`footnt_typ` is 'N'.
                    if row['footnt_no'] == '':
                        row['footnt_no'] = 1
                    if row['footnt_typ'] not in (FOOTNOTE_DESC, FOOTNOTE_MEAS, FOOTNOTE_NUTR):
                        row['footnt_typ'] = FOOTNOTE_NUTR

                    if row.get('nutr_no'):
                        nutrient = Nutrient.objects.get(number=int(row['nutr_no']))
                    else:
                        nutrient = None

                    try:
                        footnote = Footnote.objects.get(
                            food=Food.objects.get(ndb_number=int(row['ndb_no'])),
                            number=int(row['footnt_no']),
                            nutrient=nutrient
                        )
                        total_updated += 1
                    except Footnote.DoesNotExist:
                        footnote = Footnote(
                            food=Food.objects.get(ndb_number=int(row['ndb_no'])),
                            number=int(row['footnt_no']),
                            nutrient=nutrient
                        )
                        total_created += 1
                        created = True

                    footnote.type = row['footnt_typ']
                    footnote.text = row['footnt_txt']
                    footnote.save()

                    if created:
                        logging.debug('Created %s' % footnote)
                    else:
                        logging.debug('Updated %s' % footnote)

    logging.info('Created %d new footnotes' % total_created)
    logging.info('Updated %d footnotes' % total_updated)


def create_update_data_sources(data, db_alias):
    total_created = 0
    total_updated = 0

    logging.info('Processing %d data sources' % len(data))

    from usda_mongo.models import DataSource

    with switch_db(DataSource, db_alias) as DataSource:

        for row in csv.DictReader(
            data, fieldnames=(
                'datasrc_id', 'authors', 'title', 'year', 'journal',
                'vol_city', 'issue_state', 'start_page', 'end_page'
            ),
            delimiter='^', quotechar='~'
        ):
            created = False

            try:
                data_source = DataSource.objects.get(id=row['datasrc_id'])
                total_updated += 1
            except DataSource.DoesNotExist:
                data_source = DataSource(id=row['datasrc_id'])
                total_created += 1
                created = True

            data_source.description = row.get('authors')
            data_source.title = row.get('title')
            if row.get('year'):
                data_source.year = int(row['year'])
            data_source.journal = row.get('journal')
            data_source.volume_or_city = row.get('vol_city')
            data_source.issue_or_state = row.get('issue_state')
            if row.get('start_page'):
                data_source.start_page = row.get('start_page')
            if row.get('end_page'):
                data_source.end_page = row.get('end_page')
            data_source.save()

            if created:
                logging.debug('Created %s' % data_source)
            else:
                logging.debug('Updated %s' % data_source)

    logging.info('Created %d new data sources' % total_created)
    logging.info('Updated %d data sources' % total_updated)


def create_update_derivations(data, db_alias):
    total_created = 0
    total_updated = 0

    logging.info('Processing %d data derivations' % len(data))

    from usda_mongo.models import DataDerivation

    with switch_db(DataDerivation, db_alias) as DataDerivation:

        for row in csv.DictReader(
            data, fieldnames=('deriv_cd', 'deriv_desc'),
            delimiter='^', quotechar='~'
        ):
            created = False

            try:
                derivation = DataDerivation.objects.get(code=row['deriv_cd'])
                total_updated += 1
            except DataDerivation.DoesNotExist:
                derivation = DataDerivation(code=row['deriv_cd'])
                total_created += 1
                created = True

            # SR22 defines `deriv_desc` as being a maximum length of 120 characters,
            # however, there is at least one instance where `deriv_desc` is greater
            # than this max.  To deal with this, truncate to 120 characters.
            derivation.description = row['deriv_desc'][:120]
            derivation.save()

            if created:
                logging.debug('Created %s' % derivation)
            else:
                logging.debug('Updated %s' % derivation)

    logging.info('Created %d new derivations' % total_created)
    logging.info('Updated %d derivations' % total_updated)


def create_update_sources(data, db_alias):
    total_created = 0
    total_updated = 0

    logging.info('Processing %d sources' % len(data))

    from usda_mongo.models import Source

    with switch_db(Source, db_alias) as Source:

        for row in csv.DictReader(
            data, fieldnames=('src_cd', 'srccd_desc'),
            delimiter='^', quotechar='~'
        ):
            created = False

            try:
                source = Source.objects.get(code=int(row['src_cd']))
                total_updated += 1
            except Source.DoesNotExist:
                source = Source(code=int(row['src_cd']))
                total_created += 1
                created = True

            source.description = row['srccd_desc']
            source.save()

            if created:
                logging.debug('Created %s' % source)
            else:
                logging.debug('Updated %s' % source)

    logging.info('Created %d new sources' % total_created)
    logging.info('Updated %d sources' % total_updated)


def create_update_nutrient_data(data, db_alias):
    total_created = 0
    total_updated = 0

    logging.info('Processing %d nutrient data items' % len(data))

    from usda_mongo.models import NutrientData, Nutrient, Food, DataDerivation, Source

    with switch_db(NutrientData, db_alias) as NutrientData:
        with switch_db(Nutrient, db_alias) as Nutrient:
            with switch_db(Food, db_alias) as Food:
                with switch_db(DataDerivation, db_alias) as DataDerivation:
                    with switch_db(Source, db_alias) as Source:

                        for start in range(0, len(data), NUTRIENT_DATA_STEP):
                            end = min(start + NUTRIENT_DATA_STEP, len(data))

                            for row in csv.DictReader(
                                data[start:end], fieldnames=(
                                    'ndb_no', 'nutr_no', 'nutr_val', 'num_data_pts', 'std_error',
                                    'src_cd', 'deriv_cd', 'ref_ndb_no', 'add_nutr_mark', 'num_studies',
                                    'min', 'max', 'df', 'low_eb', 'up_eb', 'stat_cmt', 'cc'
                                ),
                                delimiter='^', quotechar='~'
                            ):
                                created = False

                                try:
                                    nutrient_data = NutrientData.objects.get(
                                        food=Food.objects.get(ndb_number=int(row['ndb_no'])),
                                        nutrient=Nutrient.objects.get(number=int(row['nutr_no']))
                                    )
                                    total_updated += 1
                                except NutrientData.DoesNotExist:
                                    nutrient_data = NutrientData(
                                        food=Food.objects.get(ndb_number=int(row['ndb_no'])),
                                        nutrient=Nutrient.objects.get(number=int(row['nutr_no']))
                                    )
                                    total_created += 1
                                    created = True

                                nutrient_data.nutrient_value = float(row['nutr_val'])
                                nutrient_data.data_points = int(row['num_data_pts'])
                                if row.get('std_error'):
                                    nutrient_data.standard_error = float(row['std_error'])
                                if row.get('deriv_cd'):
                                    nutrient_data.data_derivation = DataDerivation.objects.get(code=row['deriv_cd'])
                                if row.get('ref_ndb_no'):
                                    nutrient_data.reference_nbd_number = int(row['ref_ndb_no'])
                                if row.get('add_nutr_mark'):
                                    nutrient_data.added_nutrient = row['add_nutr_mark'] == 'Y'
                                if row.get('num_studies'):
                                    nutrient_data.number_of_studies = int(row['num_studies'])
                                if row.get('min'):
                                    nutrient_data.minimum = float(row['min'])
                                if row.get('max'):
                                    nutrient_data.maximum = float(row['max'])
                                if row.get('df'):
                                    nutrient_data.degrees_of_freedom = int(row['df'])
                                if row.get('low_eb'):
                                    nutrient_data.lower_error_bound = float(row['low_eb'])
                                if row.get('up_eb'):
                                    nutrient_data.upper_error_bound = float(row['up_eb'])
                                nutrient_data.statistical_comments = row.get('stat_cmt')
                                nutrient_data.confidence_code = row.get('cc')
                                nutrient_data.save()

                                if row.get('src_cd'):
                                    nutrient_data.source.append(Source.objects.get(code=row['src_cd']))
                                    nutrient_data.save()

                                if created:
                                    logging.debug('Created %s' % nutrient_data)
                                else:
                                    logging.debug('Updated %s' % nutrient_data)

        #reset_queries() # Reset DB connection to avoid using all available RAM

    logging.info('Created %d new nutrient data' % total_created)
    logging.info('Updated %d nutrient data' % total_updated)
